@@ -14,7 +14,6 @@ import (
 	"math/rand"
 	"os"
 	"runtime"
-	"sync"
 	"testing"
 )
 
@@ -143,6 +142,7 @@ func testOpts() [][]Option {
 		{WithMaxGoroutines(5000), WithMinSplitSize(500000), withSSSE3(false), withAVX2(false), withAVX512(false)},
 		{WithMaxGoroutines(1), WithMinSplitSize(500000), withSSSE3(false), withAVX2(false), withAVX512(false)},
 		{WithAutoGoroutines(50000), WithMinSplitSize(500)},
+		{WithInversionCache(false)},
 	}
 	for _, o := range opts[:] {
 		if defaultOptions.useSSSE3 {
@@ -180,7 +180,9 @@ func TestEncoding(t *testing.T) {
 
 // matrix sizes to test.
 // note that par1 matric will fail on some combinations.
-var testSizes = [][2]int{{1, 1}, {1, 2}, {3, 3}, {3, 1}, {5, 3}, {8, 4}, {10, 30}, {12, 10}, {14, 7}, {41, 17}, {49, 1}}
+var testSizes = [][2]int{
+	{1, 0}, {3, 0}, {5, 0}, {8, 0}, {10, 0}, {12, 0}, {14, 0}, {41, 0}, {49, 0},
+	{1, 1}, {1, 2}, {3, 3}, {3, 1}, {5, 3}, {8, 4}, {10, 30}, {12, 10}, {14, 7}, {41, 17}, {49, 1}}
 var testDataSizes = []int{10, 100, 1000, 10001, 100003, 1000055}
 var testDataSizesShort = []int{10, 10001, 100003}
 
@@ -220,6 +222,22 @@ func testEncoding(t *testing.T, o ...Option) {
 					if !ok {
 						t.Fatal("Verification failed")
 					}
+
+					if parity == 0 {
+						// Check that Reconstruct and ReconstructData do nothing
+						err = r.ReconstructData(shards)
+						if err != nil {
+							t.Fatal(err)
+						}
+						err = r.Reconstruct(shards)
+						if err != nil {
+							t.Fatal(err)
+						}
+
+						// Skip integrity checks
+						return
+					}
+
 					// Delete one in data
 					idx := rng.Intn(data)
 					want := shards[idx]
@@ -646,7 +664,8 @@ func testVerify(t *testing.T, o ...Option) {
 		t.Fatal(err)
 	}
 	if !ok {
-		t.Fatal("Verification failed")
+		t.Error("Verification failed")
+		return
 	}
 
 	// Put in random data. Verification should fail
@@ -761,7 +780,7 @@ func benchmarkEncode(b *testing.B, dataShards, parityShards, shardSize int) {
 		fillRandom(shards[s])
 	}
 
-	b.SetBytes(int64(shardSize * dataShards))
+	b.SetBytes(int64(shardSize * (dataShards + parityShards)))
 	b.ResetTimer()
 	for i := 0; i < b.N; i++ {
 		err = r.Encode(shards)
@@ -771,6 +790,10 @@ func benchmarkEncode(b *testing.B, dataShards, parityShards, shardSize int) {
 	}
 }
 
+func BenchmarkEncode2x1x1M(b *testing.B) {
+	benchmarkEncode(b, 2, 1, 1024*1024)
+}
+
 func BenchmarkEncode10x2x10000(b *testing.B) {
 	benchmarkEncode(b, 10, 2, 10000)
 }
@@ -853,7 +876,7 @@ func benchmarkVerify(b *testing.B, dataShards, parityShards, shardSize int) {
 		b.Fatal(err)
 	}
 
-	b.SetBytes(int64(shardSize * dataShards))
+	b.SetBytes(int64(shardSize * (dataShards + parityShards)))
 	b.ResetTimer()
 	for i := 0; i < b.N; i++ {
 		_, err = r.Verify(shards)
@@ -925,7 +948,7 @@ func benchmarkReconstruct(b *testing.B, dataShards, parityShards, shardSize int)
 		b.Fatal(err)
 	}
 
-	b.SetBytes(int64(shardSize * dataShards))
+	b.SetBytes(int64(shardSize * (dataShards + parityShards)))
 	b.ResetTimer()
 	for i := 0; i < b.N; i++ {
 		corruptRandom(shards, dataShards, parityShards)
@@ -999,7 +1022,7 @@ func benchmarkReconstructData(b *testing.B, dataShards, parityShards, shardSize
 		b.Fatal(err)
 	}
 
-	b.SetBytes(int64(shardSize * dataShards))
+	b.SetBytes(int64(shardSize * (dataShards + parityShards)))
 	b.ResetTimer()
 	for i := 0; i < b.N; i++ {
 		corruptRandomData(shards, dataShards, parityShards)
@@ -1052,7 +1075,7 @@ func benchmarkReconstructP(b *testing.B, dataShards, parityShards, shardSize int
 		b.Fatal(err)
 	}
 
-	b.SetBytes(int64(shardSize * dataShards))
+	b.SetBytes(int64(shardSize * (dataShards + parityShards)))
 	b.ResetTimer()
 
 	b.RunParallel(func(pb *testing.PB) {
@@ -1069,7 +1092,7 @@ func benchmarkReconstructP(b *testing.B, dataShards, parityShards, shardSize int
 		if err != nil {
 			b.Fatal(err)
 		}
-
+		b.ResetTimer()
 		for pb.Next() {
 			corruptRandom(shards, dataShards, parityShards)
 
@@ -1240,12 +1263,11 @@ func TestStandardMatrices(t *testing.T) {
 		// Runtime ~15s.
 		t.Skip("Skipping slow matrix check")
 	}
-	var wg sync.WaitGroup
-	wg.Add(256 - 1)
 	for i := 1; i < 256; i++ {
-		go func(i int) {
+		i := i
+		t.Run(fmt.Sprintf("x%d", i), func(t *testing.T) {
+			t.Parallel()
 			// i == n.o. datashards
-			defer wg.Done()
 			var shards = make([][]byte, 255)
 			for p := range shards {
 				v := byte(i)
@@ -1293,9 +1315,8 @@ func TestStandardMatrices(t *testing.T) {
 					}
 				}
 			}
-		}(i)
+		})
 	}
-	wg.Wait()
 }
 
 func TestCauchyMatrices(t *testing.T) {
@@ -1303,12 +1324,10 @@ func TestCauchyMatrices(t *testing.T) {
 		// Runtime ~15s.
 		t.Skip("Skipping slow matrix check")
 	}
-	var wg sync.WaitGroup
-	wg.Add(256 - 1)
 	for i := 1; i < 256; i++ {
-		go func(i int) {
-			// i == n.o. datashards
-			defer wg.Done()
+		i := i
+		t.Run(fmt.Sprintf("x%d", i), func(t *testing.T) {
+			t.Parallel()
 			var shards = make([][]byte, 255)
 			for p := range shards {
 				v := byte(i)
@@ -1356,9 +1375,8 @@ func TestCauchyMatrices(t *testing.T) {
 					}
 				}
 			}
-		}(i)
+		})
 	}
-	wg.Wait()
 }
 
 func TestPar1Matrices(t *testing.T) {
@@ -1366,12 +1384,10 @@ func TestPar1Matrices(t *testing.T) {
 		// Runtime ~15s.
 		t.Skip("Skipping slow matrix check")
 	}
-	var wg sync.WaitGroup
-	wg.Add(256 - 1)
 	for i := 1; i < 256; i++ {
-		go func(i int) {
-			// i == n.o. datashards
-			defer wg.Done()
+		i := i
+		t.Run(fmt.Sprintf("x%d", i), func(t *testing.T) {
+			t.Parallel()
 			var shards = make([][]byte, 255)
 			for p := range shards {
 				v := byte(i)
@@ -1428,9 +1444,8 @@ func TestPar1Matrices(t *testing.T) {
 					}
 				}
 			}
-		}(i)
+		})
 	}
-	wg.Wait()
 }
 
 func TestNew(t *testing.T) {
@@ -1441,10 +1456,12 @@ func TestNew(t *testing.T) {
 		{127, 127, nil},
 		{128, 128, nil},
 		{255, 1, nil},
+		{255, 0, nil},
+		{1, 0, nil},
 		{256, 256, ErrMaxShardNum},
 
 		{0, 1, ErrInvShardNum},
-		{1, 0, ErrInvShardNum},
+		{1, -1, ErrInvShardNum},
 		{256, 1, ErrMaxShardNum},
 
 		// overflow causes r.Shards to be negative
@@ -1529,7 +1546,7 @@ func benchmarkParallel(b *testing.B, dataShards, parityShards, shardSize int) {
 		shardsCh <- shards
 	}
 
-	b.SetBytes(int64(shardSize * dataShards))
+	b.SetBytes(int64(shardSize * (dataShards + parityShards)))
 	b.SetParallelism(c)
 	b.ReportAllocs()
 	b.ResetTimer()
